/* xscreensaver, Copyright (c) 2026 Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

#ifndef __EARTH_H__
#define __EARTH_H__

extern const unsigned char earth_png[];       extern int earth_png_size;
extern const unsigned char earth_night_png[]; extern int earth_night_png_size;
extern const unsigned char earth_flat_png[];  extern int earth_flat_png_size;
extern const unsigned char earth_water_png[]; extern int earth_water_png_size;

#endif /* __EARTH_H__ */
